/*
 * maths_fix16.h
 *
 * OBSERVE that lib_dsp dsp_math.h + dsp_math.xc have all that's needed here I assume
 *
 *  Created on: 5. nov. 2021
 *      Author: oyvindteig
 */

/*
From lib_dsp.pdf:
    The xCORE-200 has a single cycle 32x32->64 bit multiply/accumulate unit,
    single cycle double-word load and store, dual issue instruction execution,
    and other instruction set enhancements. These features make xCORE-200 an
    efficient platform for executing digital signal processing algorithms.

However, division/divide do have instructions, but they are not single cycle.
See lib_dsp dsp_math_divide which uses an instruction for long unsigned division

I have not found a list of instructions and cycle count but here is some more info:

Assembly Programming Manual
Publication Date: 2015/3/27 Document Number: X9432B

divs  d, x, y          Signed division
divu  d, x, y          Unsigned division
ldivu d, e, v, x, y    Long unsigned divide
lmul  d, e, x, y, v, w Long multiply
maccs d, e, x, y       Mulitply and accumulate signed
maccu d, e, x, y       Multiply and accumulate unsigned
mul   d, x, y          Multiply

Cycle count: https://www.xcore.com/viewtopic.php?f=3&t=2067
64 bit math operations? Clock cycles per operation?
    From 2013 before XS2
    You might be able to look at the disassembly of the code to see how many instructions that line translates in to,
    but if you want to know how fast it with optimizations on you can create a timer in xc.
    They count by default at 100MHz so if you create two variables and grab the timers value before and after the
    multiply operation and subtract the two you get how many cycles it tock to get there.(Might be a good idea to
    copy paste 10 multiplys and measure that so you get a more accurate reading)

Also here: https://www.xcore.com/viewtopic.php?t=5775
LDIVU questions
    And yes, as I understand it, there is only one divide unit per tile, so everything on that tile has to queue up to use it.
    When a division has to be done, I assume 32 * n worst case cycles, n being the number of threads which might use the divider.

Also https://www.xcore.com/viewtopic.php?f=26&t=2475
How do I target the LDIV instruction?
    LDIVU instruction only works if the division fits in a 32-bit word, that is, if the higher word of the double word input
    is less than the divisor. This operation is intended to be used for the implementation of long division.
*/

#ifndef MATHS_FIX16_H_
    #define MATHS_FIX16_H_

    // Fixed point         https://en.wikipedia.org/wiki/Fixed-point_arithmetic
    // See                 https://schaumont.dyn.wpi.edu/ece4703b20/lecture6.html
    // "fix16" naming from https://en.wikipedia.org/wiki/Libfixmath
    //
    // Decimal point (binary point) is between bit16 and bit15

    #define FIX16_UNITY_BITPOS 16                      // #        sign bit if signed int
    #define FIX16_UNITY        (1<<FIX16_UNITY_BITPOS) // 1.0000 (65565)
    #define FIX16_NEG_UNITY    (-1)                    // 11111111 all 1 as 2's complemenet

    typedef uint32_t ufix16_t;
    typedef int32_t  sfix16_t;

    // Based on libfixmath:
    /* Subtraction and addition no overflow detection. */

    // Not used:
    // OBSERVE that lib_dsp dsp_math.h + dsp_math.xc have all that's needed here I assume

    static inline sfix16_t fix16_add (const sfix16_t inArg0, const sfix16_t inArg1) { return (inArg0 + inArg1); }
    static inline sfix16_t fix16_sub (const sfix16_t inArg0, const sfix16_t inArg1) { return (inArg0 - inArg1); }

#else
    #error NESTED .h FILE
#endif
