/*
 * mics_in.h
 *
 *  Created on: 22. apr. 2021
 *      Author: teig
 *
 *      Needed include files, see bottom of this file
 */

#ifndef MICS_IN_HEADSET_OUT_H_
    #define MICS_IN_HEADSET_OUT_H_

    // =============
    // MICS_IN PART
    // =============

    #ifndef MIC_ARRAY_WORD_LENGTH_SHORT
        #error MIC_ARRAY_WORD_LENGTH_SHORT not defined
    #else
        // in mic_array_frame.h in lib_mic_array
        //
        // From "mic_array_conf.h" in lib_mic_array user guide PDF
        // MIC_ARRAY_WORD_LENGTH_SHORT
        //     If this define is set to non-zero then this configures the output word length to be a 16 bit
        //     short otherwise its left as 32 bit word length output. All internal processing will be done at
        //     32 bits, only during the write to frame memory will the truncation happen.
        // This is done in mic_array_init_time_domain_frame and mic_array_get_next_time_domain_frame
        //
        #if (MIC_ARRAY_WORD_LENGTH_SHORT==0) // THIS
            typedef int32_t mic_sample_t; // Data in 16 first LSB bits, ..
            #define MIC_ARRAY_WORD_DB(arg) int32_dB(arg)
            #define MIC_ARRAY_WORD_MIN INT_MIN
            #define MIC_ARRAY_WORD_MAX INT_MAX
        #elif (MIC_ARRAY_WORD_LENGTH_SHORT==1)
            typedef int16_t mic_sample_t; // ..here these 16 bits are all that's present
            #define MIC_ARRAY_WORD_DB(arg) int16_dB(arg)
            #define MIC_ARRAY_WORD_MIN SHRT_MIN
            #define MIC_ARRAY_WORD_MAX SHRT_MAX
        #endif
    #endif

    typedef int      mic_data_t; // Must be int since mic_array_decimator_config_t has int as field 2
    typedef unsigned delay_mics_num_samples_t;

    #define MAX_DELAY_SAMPLE_CNT 16
    #define NUM_PHYSICAL_MICS    (MIC_ARRAY_NUM_MICS - 1) // 7
    #define NUM_CIRCULAR_MICS    (NUM_PHYSICAL_MICS  - 1) // 6

    typedef unsigned mics_direction_t; // [NUM_CIRCULAR_MICS] = [0..(NUM_CIRCULAR_MICS-1)]

    typedef struct {
        bool             allow;
        mics_direction_t mics_direction;
    } directive_mics_t;

    // Based on the spreadsheet mic_array_das_beamformer_calcs.xls,
    // which can be found in the root directory of this app
    #define DELAY_MICS_HEAD_DEFAULT 0
    //
    #define DELAY_SAMPLE_COUNT_OF_0                       43
    #define DELAY_SAMPLE_COUNT_ONE_METER_THIRTY_DEGREES {  0, 3, 8, 11, 8, 3 } // NUM_CIRCULAR_MICS
    //
    #define START_CIRCULAR_MICS 3 // Assume this is the meaning in AN00219 in formula in set_dir

    #ifndef MIC_ARRAY_NUM_MICS
        #error mic_array_conf.h not seen
    #endif

    #define NUM_4MICS_DECIMATOR_CHANS 4 // Simply because mic_array_decimate_to_pcm_4ch takes 4 mics
    #define NUM_4MIC_DECIMATORS       (MIC_ARRAY_NUM_MICS / NUM_4MICS_DECIMATOR_CHANS) // 8 channels requires 2 decimators:
    //                                // 2/1 when MIC_ARRAY_NUM_MICS == 8/4
    #define FRAME_BUFFER_COUNT        2 // AN00220/AN00219: The minimum of 2 will suffice for these examples (when DECIMATOR_NO_FRAME_OVERLAP)

    #define NUM_DATA_X (NUM_4MIC_DECIMATORS         * NUM_4MICS_DECIMATOR_CHANS) // 8/4 when MIC_ARRAY_NUM_MICS == 8/4
    #define NUM_DATA_Y (THIRD_STAGE_COEFS_PER_STAGE * DECIMATION_FACTOR)        // 192 = 32*6

    // ================
    // HEADSET_OUT PART
    // ================

    typedef ufix16_t headset_gain_t;

    #define _HEADSET_GAIN_UNITY_BITPOS FIX16_UNITY_BITPOS
    #define _HEADSET_GAIN_UNITY        (1 << _HEADSET_GAIN_UNITY_BITPOS)

    typedef enum do_headset_gain_e {
        do_headset_gain_zero,
        do_headset_gain_default,
        do_headset_gain_max,
        do_headset_gain_increase,
        do_headset_gain_decrease
    } do_headset_gain_e;

    // Observe that AN00219 and mic_array_conf.h MIC_ARRAY_FIXED_GAIN work with shifting
    // bits (ie. div/mul by 2,4,8,16 etc), not a dB scale as I have chosen:
    //
    //     https://en.wikipedia.org/wiki/Half-power_point
    //     https://en.wikipedia.org/wiki/Decibel and https://www.rapidtables.com/electric/decibel.html
    //
    // Amplitude ratio 20*log10(Vin/Vout)
    //
    #define NUM_AMPLITUDE_STEPS        11 // 22
    #define _IOF_HEADSET_GAIN_DB_OFF    0
    #define _IOF_HEADSET_GAIN_DB_UNITY  4 // 11
    #define _IOF_HEADSET_GAIN_DB_MAX   (NUM_AMPLITUDE_STEPS-1)
    //
    #define AMPLITUDE_FIX16_UNITY_FACTORS_TABLE {  \
            0,                 /* OFF    */ \
         /* _HEADSET_GAIN_UNITY / 1000,          -80 dB */ \
         /* _HEADSET_GAIN_UNITY *    0.00316,    -50 dB */ \
         /* _HEADSET_GAIN_UNITY /  100,          -40 dB */ \
         /* _HEADSET_GAIN_UNITY *    0.032,      -30 dB */ \
         /* _HEADSET_GAIN_UNITY /   10,          -20 dB */ \
            _HEADSET_GAIN_UNITY *    0.316,   /* -10 dB */ \
            _HEADSET_GAIN_UNITY *    0.501,   /*  -6 dB (HALF 0.5 is -6.0206 dB) */ \
            _HEADSET_GAIN_UNITY *    0.708,   /*  -3 dB (1/2     is -3.0103 dB) */ \
         /* _HEADSET_GAIN_UNITY *    0.794,       -2 dB */ \
         /* _HEADSET_GAIN_UNITY *    0.891,       -1 dB */ \
            _HEADSET_GAIN_UNITY *    1,       /*   0 dB [_IOF_HEADSET_GAIN_DB_UNITY] */ \
         /* _HEADSET_GAIN_UNITY *    1.122,        1 dB */ \
         /* _HEADSET_GAIN_UNITY *    1.259,        2 dB */ \
            _HEADSET_GAIN_UNITY *    1.413,   /*   3 dB (2       is 3.0103 dB ) */ \
            _HEADSET_GAIN_UNITY *    1.995,   /*   6 dB (DOUBLE 2 is 6.0206 dB) */ \
            _HEADSET_GAIN_UNITY *    3.162,   /*  10 dB */ \
            _HEADSET_GAIN_UNITY *   10,       /*  20 dB */ \
            _HEADSET_GAIN_UNITY *   31.623,   /*  30 dB */ \
            _HEADSET_GAIN_UNITY *  100,       /*  40 dB */ \
         /* _HEADSET_GAIN_UNITY *  316.228,       50 dB */ \
         /* _HEADSET_GAIN_UNITY * 1000            60 dB */ \
    };
    #define IOF_HEADSET_GAIN_DB_DEFAULT _IOF_HEADSET_GAIN_DB_UNITY
    #define IOF_HEADSET_GAIN_DB_MAX     _IOF_HEADSET_GAIN_DB_MAX
    #define IOF_HEADSET_GAIN_DB_OFF     _IOF_HEADSET_GAIN_DB_OFF
    //
    #define HEADSET_GAIN_DEFAULT        _HEADSET_GAIN_UNITY
    #define HEADSET_GAIN_BITPOS_DEFAULT _HEADSET_GAIN_UNITY_BITPOS
    #define HEADSET_GAIN_OFF_DEFAULT    false

    #define AMPLITUDE_DB_TABLE \
    { \
            -99, /* OFF */ \
            -10, \
             -6, \
             -3, \
              0, /* [_IOF_HEADSET_GAIN_DB_UNITY] */ \
              3, \
              6, \
             10, \
             20, \
             30, \
             40, \
    };

    typedef struct {
        unsigned        iof_gain;
        bool            gain_off;
        headset_gain_t  gain;
        mic_data_t      amplitude_fix16_unity_factors_table [NUM_AMPLITUDE_STEPS];
        signed          amplitude_db_table                  [NUM_AMPLITUDE_STEPS];
    } headset_gain_context_t;

    void set_headset_gain (
            const do_headset_gain_e do_headset_gain,
            headset_gain_context_t      &ctx);

    #define I2C_ADDRESS_OF_DAC 0x4A // CS43L21, see page 5 of mic_array board manual
    //
    typedef enum i2c_reg_addr_dac_headphones_cs43l21_e {
        CS41L21_DAC_REG_01_CHIP_ID_AND_REVISION                      = 0x01, // USED
        CS41L21_DAC_REG_02_POWER_CONTROL_1                           = 0x02, // USED
        CS41L21_DAC_REG_03_SPEED_CONTROL                             = 0x03, // USED
        CS41L21_DAC_REG_04_INTERFACE_CONTROL                         = 0x04,
        CS41L21_DAC_REG_08_DAC_OUTPUT_CONTROL                        = 0x08,
        CS41L21_DAC_REG_09_DAC_CONTROL                               = 0x09,
        CS41L21_DAC_REG_10_MIXER_VOLUME_CONTROL_PCMA                 = 0x10, // USED
        CS41L21_DAC_REG_11_MIXER_VOLUME_CONTROL_PCMB                 = 0x11,
        CS41L21_DAC_REG_12_BEEP_FREQUENCY_AND_TIMING_CONFIGURATION   = 0x12,
        CS41L21_DAC_REG_13_BEEP_OFF_TIME_AND_VOLUME                  = 0x13,
        CS41L21_DAC_REG_14_BEEP_CONFIGURATION_AND_TONE_CONFIGURATION = 0x14,
        CS41L21_DAC_REG_15_TONE_CONTROL                              = 0x15,
        CS41L21_DAC_REG_16_VOLUME_CONTROL_AOUTA                      = 0x16,
        CS41L21_DAC_REG_17_VOLUME_CONTROL_AOUTB                      = 0x17,
        CS41L21_DAC_REG_18_PCM_CHANNEL_MIXER                         = 0x18,
        CS41L21_DAC_REG_19_LIMITER_THRESHOLD_SZC_DISABLE             = 0x19,
        CS41L21_DAC_REG_1A_LIMITER_RELEASE_RATE_REGISTER             = 0x1A,
        CS41L21_DAC_REG_1B_LIMITER_ATTACK_RATE_REGISTER              = 0x1B,
        CS41L21_DAC_REG_20_STATUS                                    = 0x20,
        CS41L21_DAC_REG_21_CHARGE_PUMP_FREQUENCY                     = 0x21
    } i2c_reg_addr_dac_headphones_cs43l21_e;

    #define I2C_ADDRESS_OF_PLL 0x4E // CS2100-CP, see page 5 of mic_array board manual
    //
    typedef enum i2c_reg_addr_pll_cs2100cp_e {
        CS2100CP_PLL_REG_01_DEVICE_ID_AND_REVISION   = 0x01,
        CS2100CP_PLL_REG_02_DEVICE_CONTROL           = 0x02,
        CS2100CP_PLL_REG_03_DEVICE_CONFIGURATION_1   = 0x03,
        CS2100CP_PLL_REG_05_GLOBAL_CONFIGURATION     = 0x05,
        CS2100CP_PLL_REG_06_RATIO                    = 0x06,
        CS2100CP_PLL_REG_07_RATIO                    = 0x07,
        CS2100CP_PLL_REG_08_RATIO                    = 0x08,
        CS2100CP_PLL_REG_09_RATIO                    = 0x09,
        CS2100CP_PLL_REG_16_FUNCTION_CONFIGURATION_1 = 0x16,
        CS2100CP_PLL_REG_17_FUNCTION_CONFIGURATION_2 = 0x17,
        CS2100CP_PLL_REG_1E_FUNCTION_CONFIGURATION_3 = 0x1E
    } i2c_reg_addr_pll_cs2100cp_e;

    [[distributable]]
    void i2s_handler (
            server  i2s_callback_if i2s, // From i2s_master0
            client  i2c_master_if   i2c,
            port    p_rst_shared,
            chanend c_audio);

    // =============
    // COMMON IN-OUT
    // =============

    typedef struct mic_result_t {
        unsigned     cnt;
        mic_sample_t max;
        mic_sample_t min;
    } mic_result_t;

    typedef struct menu_result_t {
        bool             headset_gain_updated;
        bool             headset_gain_off;
        headset_gain_t   headset_gain;
        bool             directive_mics_valid;
        directive_mics_t directive_mics;
    } menu_result_t;

    typedef enum {
        task_a,
        task_b
    } ab_src_e;

    // Between mics_in_headset_out_task_a and Handle_Beep_BRRRR_task_b
    // task_a sends important data and task_b some times adds new menu changes
    //
    typedef struct {
        ab_src_e source;
        union {
            mic_result_t  mic_result;  // spontaneous message: source is task_a
            menu_result_t menu_result; // required response:   source is task_b
        } data;
    } ch_ab_bidir_t;

    // OBSOLETED with 0203:
    typedef interface mic_result_if {
        void send_mic_data (const mic_result_t mic_result);
    } mic_result_if;

    #define MASTER_TO_PDM_CLOCK_DIVIDER 4
    #define MASTER_CLOCK_FREQUENCY      24576000
    #define PDM_CLOCK_FREQUENCY         (MASTER_CLOCK_FREQUENCY / (2 * MASTER_TO_PDM_CLOCK_DIVIDER))
    #define OUTPUT_SAMPLE_RATE          (PDM_CLOCK_FREQUENCY / (32 * DECIMATION_FACTOR))

    void mics_in_headset_out_task_a (
            streaming chanend ds_output_is_ptr_chan [NUM_4MIC_DECIMATORS], // [2/1] when MIC_ARRAY_NUM_MICS == 8/4
            mic_data_t        mics_data [NUM_DATA_X] [NUM_DATA_Y],         // [8/4] [192] when MIC_ARRAY_NUM_MICS == 8/4. Double-word aligned
            out port          p_scope, // (*)
            chanend           ch_ab_bidir, // ch_ab_bidir_t
            chanend           ch_headset_out);

            // (*) Having this as a param instead of being a global, as in the AN00220 example, yields the same
            // amount of all resources, on the byte, but the XTA timing analysis is _a_little_ worse. Strange. See files
            // "2021 04 23 09.40 build data global to example.txt"
            //     xta: /Users/teig/workspace/lib_mic_array/src/decimate_to_pcm_4ch.S:30: warning: route(4)     Pass with 4 unknowns, Num Paths: 1, Slack: 994.0 ns, Required: 2.6 us, Worst: 1.6 us, Min Core Frequency: 309 MHz
            // "2021 04 23 09.40 build data param to example.txt"                                                                                          ########                                                       #######
            //     xta: /Users/teig/workspace/lib_mic_array/src/decimate_to_pcm_4ch.S:30: warning: route(4)     Pass with 4 unknowns, Num Paths: 1, Slack: 984.0 ns, Required: 2.6 us, Worst: 1.6 us, Min Core Frequency: 311 MHz
#else
    #error NESTED .h FILE. BUT IT NEEDS:
    /* TODO: make this list for all header files
    lib_mic_array included in makefile will resulve  local "mic_array_conf.h" to be seen
    #include <i2s.h>
    #include <i2c.h>
    #include "mic_array.h"
    #include "_globals.h"
    #include "maths_fix16.h"
    */
#endif
