/*
 * AN00220_app_phase_aligned_example.xc
 *
 *  Created on: 22. apr. 2021
 *      Author: XMOS with AN00220
 */
#define INCLUDES
#ifdef INCLUDES
    #include <platform.h>
    #include <xs1.h>
    #include <string.h>
    #include <iso646.h> // not etc.
    #include <limits.h> // INT_MAX
    #include <stdint.h> // int32_t etc
    #include <xscope.h> // xScope logging
    #include <i2c.h>
    #include <i2s.h>
    #include <xassert.h>

    #include "mic_array.h"
    #include "mic_array_board_support.h"

    #include "_globals.h"
    #include "maths_fix16.h" // Needs stdint.h
    #include "mics_in_headset_out.h"

    #define DO_DEBUG_PRINT 1
    #define DO_WARNINGS    0
    #include "_print_macros_xc.h"
#endif

// ============
// MICS_IN_PART
// ============

#if (WARNINGS==1)
    #ifndef MIC_ARRAY_WORD_LENGTH_SHORT
        #warning MIC_ARRAY_WORD_LENGTH_SHORT not defined
    #else // in mic_array_frame.h lib_mic_array
        #if (MIC_ARRAY_WORD_LENGTH_SHORT==0) // THIS
            #warning int32_t bits mic array word
        #elif (MIC_ARRAY_WORD_LENGTH_SHORT==1)
            #warning int16_t bits mic array word
        #endif
    #endif

    #ifndef MIC_ARRAY_NUM_MICS
        #warning MIC_ARRAY_NUM_MICS not defined
    #elif (MIC_ARRAY_NUM_MICS == 8) // THIS
        #warning 8 mics (7 used) // Magically from mic_array_conf.h <- mic_array_frame.h
    #endif

    #ifndef USE_XSCOPE
        #error USE_XSCOPE not defined in makefile
    #elif (USE_XSCOPE == 0)
        #warning NO XSCOPE
    #elif (USE_XSCOPE==1)
        #warning USING XSCOPE
    #endif
#endif

#define IOF_MIC_0 0 // Only the center mic just now
#define IOF_BUF_0 0

void set_headset_gain (
        const do_headset_gain_e do_headset_gain,
        headset_gain_context_t      &ctx)
{

    switch (do_headset_gain) {
        case do_headset_gain_zero: {
            ctx.iof_gain = IOF_HEADSET_GAIN_DB_OFF;
        } break;
        case do_headset_gain_default: {
            const mic_data_t  amplitude_fix16_unity_factors_table_ [NUM_AMPLITUDE_STEPS] = AMPLITUDE_FIX16_UNITY_FACTORS_TABLE;
            const signed      amplitude_db_table_                  [NUM_AMPLITUDE_STEPS] = AMPLITUDE_DB_TABLE;

            // I can understand the use of memcpy when it's this much code not to use it (but I don't like memcpy)
            for (unsigned ix=0; ix < NUM_AMPLITUDE_STEPS; ix++) {
                ctx.amplitude_fix16_unity_factors_table [ix] = amplitude_fix16_unity_factors_table_ [ix];
                ctx.amplitude_db_table                  [ix] = amplitude_db_table_                  [ix];
            }
            ctx.iof_gain = IOF_HEADSET_GAIN_DB_DEFAULT;
            ctx.gain_off = HEADSET_GAIN_OFF_DEFAULT;
        } break;
        case do_headset_gain_max: {
            ctx.iof_gain = IOF_HEADSET_GAIN_DB_MAX;
        } break;
        case do_headset_gain_increase: {
            if (ctx.iof_gain < IOF_HEADSET_GAIN_DB_MAX) {
                ctx.iof_gain++; // Into IOF_HEADSET_GAIN_DB_MAX
            }
        } break;
        case do_headset_gain_decrease: {
            if (ctx.iof_gain > IOF_HEADSET_GAIN_DB_OFF) {
                ctx.iof_gain--; // Into IOF_HEADSET_GAIN_DB_OFF
            }
        } break;
    }

    ctx.gain_off = (ctx.iof_gain == IOF_HEADSET_GAIN_DB_OFF);
    ctx.gain     = ctx.amplitude_fix16_unity_factors_table[ctx.iof_gain];
}


void set_7_mics_listening_direction_delay ( // "set_dir" in AN00219
        const directive_mics_t   directive_mics,
        const unsigned           one_meter_thirty_degrees [NUM_CIRCULAR_MICS],
        delay_mics_num_samples_t delay [NUM_PHYSICAL_MICS])  // This is updated
{

    if (directive_mics.allow) {
        delay[0] = DELAY_SAMPLE_COUNT_OF_0;
        for (unsigned i=0; i< NUM_CIRCULAR_MICS; i++) { //                              3                     6                    6
            delay[i+1] = one_meter_thirty_degrees [(i - directive_mics.mics_direction + START_CIRCULAR_MICS + NUM_CIRCULAR_MICS) % NUM_CIRCULAR_MICS];
        //                                                                                                    ((8 - 1) - 1) no matter what the IDE says
        // in mics_in_headset_out.xi (shift cmd .) to make .build visible in Finder
        // delay[0] = 43;
        //     for (unsigned i=0; i< ((8 - 1) - 1); i++) {
        //        delay[i+1] = one_meter_thirty_degrees [(i - directive_mics.mics_direction + 3 + ((8 - 1) - 1)) % ((8 - 1) - 1)];

            // mics_direction=0
            //                  i=0: one_meter_thirty_degrees[ 9%6]=[3]
            //                  i=1: one_meter_thirty_degrees[10%6]=[4]
            //                  i=2: one_meter_thirty_degrees[11%6]=[5]
            //                  i=3: one_meter_thirty_degrees[12%6]=[0]
            //                  i=4: one_meter_thirty_degrees[13%6]=[1]
            //                  i=5: one_meter_thirty_degrees[12%6]=[2]
            // mics_direction=1
            //                  i=0: one_meter_thirty_degrees[ 8%6]=[2]
            //                  i=1: one_meter_thirty_degrees[ 9%6]=[3]
            //                  i=2: one_meter_thirty_degrees[10%6]=[4]
            //                  i=3: one_meter_thirty_degrees[11%6]=[5]
            //                  i=4: one_meter_thirty_degrees[12%6]=[0]
            //                  i=5: one_meter_thirty_degrees[13%6]=[1]
            // etc..
        }
    } else {
        for (unsigned i=0; i< NUM_PHYSICAL_MICS; i++) {
            delay[i] = 0;
        }
    }
}

bool mics_handler (
        const unsigned              iof_buffer,
        mic_array_frame_time_domain audio[FRAME_BUFFER_COUNT], // double-word aligned with long long first
        mic_result_t                &mic_result) {

    // Also contains metadata[2/1 when MIC_ARRAY_NUM_MICS == 8/4] with
    // unsigned sig_bits[4];  A bit mask of the absolute values of all samples in a channel. Seems to be four zeros!
    // unsigned frame_number; The frame number
    /*
    if ((audio[iof_buffer].metadata->frame_number % (SAMPLING_FREQUENCY_HZ/2)) == 0) {
        debug_print ("cnt:%u iof_buffer:%u sig_bits:%x %x %x %x\n", // formatted %8x does not work with lib_logging I think
                audio[iof_buffer].metadata->frame_number,
                iof_buffer,
                audio[iof_buffer].metadata->sig_bits[0], // masks:
                audio[iof_buffer].metadata->sig_bits[1],
                audio[iof_buffer].metadata->sig_bits[2],
                audio[iof_buffer].metadata->sig_bits[3]);
    }
    */

    // The frame index used 0 for the oldest samples and increasing indices for newer samples.
    for (unsigned iof_frame = 0; iof_frame < MIC_ARRAY_NUM_SAMPLES; iof_frame++) {

        // iof_sample is 0 for the oldest samples and increasing indices for newer samples
        // Only [0] us "current" sample (see lores_DAS_fixed in AN00219)

        mic_sample_t sample = audio[iof_buffer].data[IOF_MIC_0][iof_frame];
        XSCOPE_INT (VALUE, sample);

        if (sample > mic_result.max) {
            mic_result.max = sample;
        } else {}

        // The first time a value like 0 will be both larger than MIC_ARRAY_WORD_MIN and less than MIC_ARRAY_WORD_MAX
        // So, cannot just "else" here

        if (sample < mic_result.min) {
            mic_result.min = sample;
        } else {}
    }

    mic_result.cnt = audio[iof_buffer].metadata->frame_number;

    return ((audio[iof_buffer].metadata->frame_number % (SAMPLING_FREQUENCY_HZ/2)) == 0); // EVERY 500 ms! See scope pictures for 0104 and 0105
}

// ================
// HEADSET_OUT_PART
// ================

// https://www.xcore.com/viewtopic.php?f=8&t=7680&p=36665&hilit=AN00219&sid=fe2fd6161bedaa637ae83af815dcf522#p36665
// mon2
#define TEST_SINE_480HZ 0
//
#if (TEST_SINE_480HZ==1)
    #define SINE_TABLE_SIZE 100 // 48 kHz / 100 = 480 Hz
    const int32_t sine_table[SINE_TABLE_SIZE] = // Complete sine
    {
        0x0100da00,0x0200b000,0x02fe8100,0x03f94b00,0x04f01100,
        0x05e1da00,0x06cdb200,0x07b2aa00,0x088fdb00,0x09646600,
        0x0a2f7400,0x0af03700,0x0ba5ed00,0x0c4fde00,0x0ced5f00,
        0x0d7dd100,0x0e00a100,0x0e754b00,0x0edb5a00,0x0f326700,
        0x0f7a1800,0x0fb22700,0x0fda5b00,0x0ff28a00,0x0ffa9c00,
        0x0ff28a00,0x0fda5b00,0x0fb22700,0x0f7a1800,0x0f326700,
        0x0edb5a00,0x0e754b00,0x0e00a100,0x0d7dd100,0x0ced5f00,
        0x0c4fde00,0x0ba5ed00,0x0af03700,0x0a2f7400,0x09646600,
        0x088fdb00,0x07b2aa00,0x06cdb200,0x05e1da00,0x04f01100,
        0x03f94b00,0x02fe8100,0x0200b000,0x0100da00,0x00000000,
        0xfeff2600,0xfdff5000,0xfd017f00,0xfc06b500,0xfb0fef00,
        0xfa1e2600,0xf9324e00,0xf84d5600,0xf7702500,0xf69b9a00,
        0xf5d08c00,0xf50fc900,0xf45a1300,0xf3b02200,0xf312a100,
        0xf2822f00,0xf1ff5f00,0xf18ab500,0xf124a600,0xf0cd9900,
        0xf085e800,0xf04dd900,0xf025a500,0xf00d7600,0xf0056400,
        0xf00d7600,0xf025a500,0xf04dd900,0xf085e800,0xf0cd9900,
        0xf124a600,0xf18ab500,0xf1ff5f00,0xf2822f00,0xf312a100,
        0xf3b02200,0xf45a1300,0xf50fc900,0xf5d08c00,0xf69b9a00,
        0xf7702500,0xf84d5600,0xf9324e00,0xfa1e2600,0xfb0fef00,
        0xfc06b500,0xfd017f00,0xfdff5000,0xfeff2600,0x00000000,
    };
#endif

[[distributable]]
void i2s_handler (
        server  i2s_callback_if i2s, // From i2s_master0
        client  i2c_master_if   i2c,
        port    p_rst_shared,
        chanend ch_headset_out) // Coming from "mics_in_headset_out_task_a" having "out" semantics
{

    p_rst_shared <: 0xF; // Teig: scope: low for ages before this! With R71

    mabs_init_pll (i2c, ETH_MIC_ARRAY);

    i2c_regop_res_t res;
    // The first thing I had done when I suspected the I2C was to print out all the I2Creturns
    // i2c_regop_res_tat the init section of i2s_handler. They were allI2C_REGOP_SUCCESS,
    // even with i2c_lib 4.0.0. But then, there is no parity or CRC checks here, and what
    // would still be open would be single bit errors in the mic_data, even if the start and stop
    // conditions were successful.

    // INIT THE DAC AND PLL FOR THE HEADPHONES
    uint8_t mic_data;

    // CS41L21_DAC_REG_01_CHIP_ID_AND_REVISION
    // Bit7     6        5        4        3        2       1       0
    // Chip_ID4 Chip_ID3 Chip_ID2 Chip_ID1 Chip_ID0 Rev_ID2 Rev_ID1 Rev_ID0
    mic_data = i2c.read_reg (I2C_ADDRESS_OF_DAC, CS41L21_DAC_REG_01_CHIP_ID_AND_REVISION, res);
    // debug_print ("I2C res 1:%u ", res);

    // CS41L21_DAC_REG_02_POWER_CONTROL_1 = POWER DOWN
    // Bit7     6        5        4        3        2        1        0
    // Reserved PDN_DACB PDN_DACA Reserved Reserved Reserved Reserved PDN
    mic_data = i2c.read_reg (I2C_ADDRESS_OF_DAC, CS41L21_DAC_REG_02_POWER_CONTROL_1, res);
    // debug_print (" 2:%u ", res);
    mic_data |= 1; // Set Power Down (PDN) bit0 to 1=Enable
    res = i2c.write_reg (I2C_ADDRESS_OF_DAC, CS41L21_DAC_REG_02_POWER_CONTROL_1, mic_data); // Power down
    // debug_print (" 3:%u ", res);

    // CS41L21_DAC_REG_03_SPEED_CONTROL: Setting MCLKDIV2 high if using 24.576MHz.
    // Bit7 6      5      4       3        2        1        0
    // AUTO SPEED1 SPEED0 3-ST_SP Reserved Reserved Reserved MCLKDIV2
    mic_data = i2c.read_reg (I2C_ADDRESS_OF_DAC, CS41L21_DAC_REG_03_SPEED_CONTROL, res);
    // debug_print (" 4:%u ", res);
    mic_data |= 1; // Set MCLK Divide By 2 (MCLKDIV2) bit0 to 1=Divide by 2
    res = i2c.write_reg (I2C_ADDRESS_OF_DAC, CS41L21_DAC_REG_03_SPEED_CONTROL, mic_data);
    // debug_print (" 5:%u ", res);

    // CS41L21_DAC_REG_10_MIXER_VOLUME_CONTROL_PCMA
    // Bit7     6       5        4        3        2        1        0
    // MUTE_    PCMMIXx PCMMIXx_ PCMMIXx_ PCMMIXx_ PCMMIXx_ PCMMIXx_ PCMMIXx_
    // PCMMIXx_ VOL6    VOL5     VOL4     VOL3     VOL2     VOL1     VOL0
    // 0        1       1        1        0        0        0        0
    mic_data = 0b01110000; // 0x70 = 112. * 0.5 dB = 56 dB gain
    res = i2c.write_reg (I2C_ADDRESS_OF_DAC, CS41L21_DAC_REG_10_MIXER_VOLUME_CONTROL_PCMA, mic_data);
    // debug_print (" 6:%u ", res);

    // CS41L21_DAC_REG_02_POWER_CONTROL_1 POWER UP
    // Bit7     6        5        4        3        2        1        0
    // Reserved PDN_DACB PDN_DACA Reserved Reserved Reserved Reserved PDN
    mic_data = i2c.read_reg (I2C_ADDRESS_OF_DAC, CS41L21_DAC_REG_02_POWER_CONTROL_1, res);
    // debug_print (" 7:%u ", res);
    mic_data &= ~1; // and_eq 0xFE is clear Power Down (PDN) bit0 0=Disable
    res = i2c.write_reg (I2C_ADDRESS_OF_DAC, CS41L21_DAC_REG_02_POWER_CONTROL_1, mic_data); // Power up
    // debug_print (" 8:%u ", res); // Power up

    // CS2100CP_PLL_REG_03_DEVICE_CONFIGURATION_1 INIT THE PLL.
    // IF FOR NOTHING ELSE; GOOD FOR ENSURING WORKING CHIP!
    // Bit7     6        5        4        3        2          1          0
    // RModSel2 RModSel1 RModSel0 Reserved Reserved AuxOutSrc1 AuxOutSrc0 EnDevCfg1
    // 0        0        0        0        0        0          0          0 is init value anyhow!
    //                                              0          0            Timing Reference Clock = RefClk
    res = i2c.write_reg (I2C_ADDRESS_OF_PLL, CS2100CP_PLL_REG_03_DEVICE_CONFIGURATION_1, 0); // RefClk to AUX_AOUT
    // debug_print (" 9:%u\n", res);

    while (1) {
        select { // Coming from i2s_master as "callbacks"  (only i2s.send is a callback I think)

            case i2s.init (
                    i2s_config_t &?i2s_config,  // ? no if (!isnull(i2s_config))
                    tdm_config_t &?tdm_config): // ? no if (!isnull(tdm_config))
            {
                i2s_config.mode = I2S_MODE_LEFT_JUSTIFIED;
                i2s_config.mclk_bclk_ratio = (MASTER_CLOCK_FREQUENCY/OUTPUT_SAMPLE_RATE)/64;
            } break;

            case i2s.restart_check() -> i2s_restart_t restart: {
                restart = I2S_NO_RESTART;
            } break;

            case i2s.receive (size_t index, int32_t sample): {
                // No code
            } break;

            case i2s.send (size_t index) -> int32_t sample: {
                // When "mics_in_headset_out_task_a" sends the value here, the
                // lib_i2s task "output_word" I think (several would do i2s_i.send), has done this code:
                //     p_dout[index] <: bitrev(i2s_i.send(index*FRAME_WORDS + offset));
                // and is waiting for data coming in on "ch_headset_out" here.
                // In other word, it hangs here. I assume this is because of output
                // frequency having to be stable, controlled from lib_i2s.
                // But "mics_in_headset_out_task_a" needs to be fast enough here, to keep pace
                // From lib_i2s PDF: "sample-based versions make a callback per channel within a sample period."
                ch_headset_out :> sample; // sample is returned (or "sent") from here
            } break;
        }
    }
}

// ===========
// COMMON PART
// ===========


void mics_in_headset_out_task_a (
        streaming chanend ds_output_is_ptr_chan[NUM_4MIC_DECIMATORS],
        mic_data_t        mics_data [NUM_DATA_X] [NUM_DATA_Y], // Double-word aligned. For mic_array_decimator_config_t dc
        out port          p_scope,
        chanend           ch_ab_bidir, // ch_ab_bidir_t
        chanend           ch_headset_out) // "out" both here and in
{
    #if (TEST_SINE_480HZ==1)
        size_t iof_sine = 0;
        debug_print ("%s\n", "mics_in_headset_out_task_a started, using sine out");
    #else
        debug_print ("%s\n", "mics_in_headset_out_task_a started");
    #endif

    { // print block
        #if (MIC_ARRAY_WORD_LENGTH_SHORT==0) // THIS
            #define MIC_DATA_WORD_BITS 32
        #elif (MIC_ARRAY_WORD_LENGTH_SHORT==1)
            #define MIC_DATA_WORD_BITS 16
        #endif

        debug_print ("audio [bufs:%u]data_%u[mics:%u][samples:%u] and mic_data int[NUM_DATA_X:%u][NUM_DATA_Y:%u]\n",
                                  FRAME_BUFFER_COUNT,
                                          MIC_DATA_WORD_BITS,
                                                  MIC_ARRAY_NUM_MICS,
                                                           1<<MIC_ARRAY_MAX_FRAME_SIZE_LOG2,
                                                                                       NUM_DATA_X,    NUM_DATA_Y);
        // PRINTS:
        // audio [bufs:2]data_32[mics:4][samples:8] and mic_data int[NUM_DATA_X:4][NUM_DATA_Y:192]
    }

    unsigned                          iof_buffer;
    mic_array_decimator_conf_common_t dcc;
    mic_array_decimator_config_t      dc[NUM_4MIC_DECIMATORS];
    mic_array_frame_time_domain       audio [FRAME_BUFFER_COUNT]; // mic_array_frame_time_domain in "mic_array_frame.h". Double word aligned with long long first

    // ================================
    // Needed when DIRECTIVE_MICS_ALLOW
    //
    mic_sample_t              data_mic_samples_delayed [MAX_DELAY_SAMPLE_CNT][NUM_PHYSICAL_MICS];
    delay_mics_num_samples_t  delay_mics_num_samples [NUM_PHYSICAL_MICS];
    delay_mics_num_samples_t  delay_mics_head = DELAY_MICS_HEAD_DEFAULT;
    directive_mics_t          directive_mics;

    const unsigned one_meter_thirty_degrees_delay_sample_count [NUM_CIRCULAR_MICS] = DELAY_SAMPLE_COUNT_ONE_METER_THIRTY_DEGREES;

    directive_mics.allow          = DIRECTIVE_MICS_ALLOW;
    directive_mics.mics_direction = 0; // 0 is always first in array, name not needed

    set_7_mics_listening_direction_delay (
            directive_mics, // in
            one_meter_thirty_degrees_delay_sample_count, // in
            delay_mics_num_samples); // This got new values. First time [43,11,8,3,0,3,8]

    unsafe {
        memset (data_mic_samples_delayed, 0, MAX_DELAY_SAMPLE_CNT * NUM_PHYSICAL_MICS * sizeof (mic_sample_t));
    }
    //
    // ================================

    timer          tmr;
    time32_t       time_ticks;
    bool           scope_pin = false;
    mic_result_t   mic_result;

    headset_gain_t headset_gain     = HEADSET_GAIN_DEFAULT;
    bool           headset_gain_off = HEADSET_GAIN_OFF_DEFAULT;

    mic_result.cnt = 0;
    mic_result.max = MIC_ARRAY_WORD_MIN;
    mic_result.min = MIC_ARRAY_WORD_MAX;

    p_scope <: scope_pin;

    unsafe {
        memset (mics_data, 0, NUM_DATA_X * NUM_DATA_Y * sizeof (mic_data_t));

        // At the lib_mic_array this is a pointer called dcc inside mic_array_decimator_config_t
        // there parameterised as mic_array_decimator_config_t dc[] into the functions
        // and typically used as dc[0].dcc->buffering_type in decimator_interface.xc in lib_mic_array
        //
        // Common to all microphones:

        mic_array_decimator_conf_common_t dcc_ = {
                                               //                                 asm in decimate_to_pcm_4ch.S probably used as (in mic_array_decimate_to_pcm_4ch)
                MIC_ARRAY_MAX_FRAME_SIZE_LOG2, //     frame_size_log2             S_FRAME_SIZE_LOG2                                    Frame size log 2 is set to 0, i.e. one sample per channel will be present in each frame
                1,                             //     apply_dc_offset_removal     S_DC_OFFSET_REMOVAL_ENABLED                          DC offset elimination is turned on. With single pole IIR filter:
                                               //                                                                                      Y[n] = Y[n-1] * alpha + x[n] - x[n-1] which mutes DC
                0,                             //     index_bit_reversal          S_INDEX_BITREVERSING_ENABLED                         Index bit reversal is off, set to 1 if FFT to be used here
                0,                             // ptr windowing_function          S_WINDOWING_ENABLED                                  No windowing function is being applied
                DECIMATION_FACTOR,             //     output_decimation_factor    S_DECIMATION_FACTOR_is_now_S_THIRD_STAGE_PHASE_COUNT The decimation factor is set to 6. Added to the fixed factor of 4
                COEF_ARRAYS,                   // ptr coefs                       S_THIRD_STAGE or S_THIRD_STAGE_PHASE_COUNT           This is a pointer to an array of arrays containing the coefficients for the final stage of decimation.
                                               //                                                                                      This corresponds to a 16kHz output hence this coef array is used
                0,                             //     apply_mic_gain_compensation S_MIC_GAIN_COMP                                      Gain compensation is turned off
                FIR_GAIN_COMPENSATION,         //     fir_gain_compensation       S_FIR_GAIN_COMP                                      FIR compensation is set to the corresponding coefficients
                DECIMATOR_NO_FRAME_OVERLAP,    //     buffering_type              S_D_FRAME_NO_OVERLAPPING                             Frame overlapping is turned off (then FRAME_BUFFER_COUNT min 2)
                FRAME_BUFFER_COUNT             //     number_of_frame_buffers     ?                                                    The number of buffers in the audio array
                                               //                                                                                      The count of frames used between the decimators and the application
        };

        // Each part (of the two) is specific to the batch of 4 microphones it interfaces to:
        // mic_array_decimator_conf_common_t * unsafe dcc;
        // int * unsafe mic_data;             // The mic_data for the FIR decimator
        // int mic_gain_compensation[4];  // An array describing the relative gain compensation to apply to the microphones.
        //                                // The microphone with the least gain is defined as 0x7fffffff (INT_MAX),
        //                                // all others are given as INT_MAX*min_gain/current_mic_gain.
        // unsigned channel_count;        // The count of enabled channels (0->4).
        //
        mic_array_decimator_config_t dc_[NUM_4MIC_DECIMATORS] = {
            {
                    &dcc,                                 // ptr dcc                   Defined above
                    mics_data[0],                         // ptr mic_data                  The storage area for the output decimator
                    {INT_MAX, INT_MAX, INT_MAX, INT_MAX}, //     mic_gain_compensation Microphone gain compensation (turned off)
                    NUM_4MICS_DECIMATOR_CHANS             //     channel_count         Enabled channel count (currently must be 4)
            }
            #if (MIC_ARRAY_NUM_MICS == 8)
            ,
            {
                    &dcc,                                              // ptr dcc                   Defined above
                    mics_data[MIC_ARRAY_NUM_MICS/NUM_4MIC_DECIMATORS], // ptr mic_data                  The storage area for the output decimator
                    {INT_MAX, INT_MAX, INT_MAX, INT_MAX},              //     mic_gain_compensation Microphone gain compensation (turned off)
                    NUM_4MICS_DECIMATOR_CHANS                          //     channel_count         Enabled channel count (currently must be 4)
            }
            #endif
        };

        memcpy (&dcc, &dcc_, sizeof(dcc_));
        memcpy ( dc,   dc_,  sizeof(dc_));

    } // unsafe

    mic_array_decimator_configure    (ds_output_is_ptr_chan, NUM_4MIC_DECIMATORS, dc); // all params input
    mic_array_init_time_domain_frame (ds_output_is_ptr_chan, NUM_4MIC_DECIMATORS, iof_buffer, audio, dc);

    tmr :> time_ticks;

    while (1) {
        // =============================================================
        // See "Intended usage model" in in lib_mic_array user guide PDF
        // =============================================================
        //
        mic_array_frame_time_domain * latest_frame; // A ptr to raw audio data data (as mic_sample_t = MIC_ARRAY_WORD_LENGTH_SHORT = int16_t or int32_t)
                                                    // which is double-word aligned with a long long initially. Declared as alias at function.
                                                    // Same mic_data as in audio (below)
        latest_frame =                              // The ptr to the frame now owned by the application. That is, the most recently written samples
            mic_array_get_next_time_domain_frame (  // in lib_mic_array, file decimator_interface.xc
                    ds_output_is_ptr_chan,          // The channels used to transfer pointers between the application and the mic_array_decimate_to_pcm_4ch() tasks
                    NUM_4MIC_DECIMATORS,            // The count of mic_array_decimate_to_pcm_4ch() tasks.
                    iof_buffer,                     // return: The buffer index (Used internally)
                    audio,                          // return: An array of mic_array_frame_time_domain audio frames.
                                                    // All frame types contain a two-dimensional mic_data array
                                                    // Same data as in latest_frame (above)
                    dc);                            // The array containing the decimator configuration for each decimator
        //
        // Uses these, found in in xs1.h:
        //     void schkct(streaming chanend c, unsigned char val);
        //     void soutct(streaming chanend c, unsigned char val);
        if (mics_handler (iof_buffer, audio, mic_result)) {
            ch_ab_bidir_t ab_bidir_context; // Here: limits scope of struct with union

            // SET OUT MESSAGE
            //
            ab_bidir_context.source = task_a;
            ab_bidir_context.data.mic_result = mic_result;

            // SEND OUT MESSAGE
            // Make OUT/IN as atomic as possible so that it will block at short time as possible
            //
            ch_ab_bidir <: ab_bidir_context;

            // RECEIVE RETURN MESSAGE
            //
            ch_ab_bidir :> ab_bidir_context;

            // ANALYSE RETURN MESSAGE
            //
            xassert (ab_bidir_context.source == task_b);

            if (ab_bidir_context.data.menu_result.headset_gain_updated) {
                headset_gain     = ab_bidir_context.data.menu_result.headset_gain;
                headset_gain_off = ab_bidir_context.data.menu_result.headset_gain_off;
            } else {}

            // OTHER
            //
            mic_result.max = MIC_ARRAY_WORD_MIN;
            mic_result.min = MIC_ARRAY_WORD_MAX;
        } else {}

        scope_pin = not scope_pin;
        p_scope <: scope_pin; // One edge per 16 kHz = 8 kHz

        if (headset_gain_off) {
            // Do nothing here
        } else {
            int output_to_dac = 0;

            // From AN00219 (also see "Accessing the samples" in the lib_mic_array PDF)
            //
            // Copy the latest_frame sample to the delay_mics_num_samples iof_buffer
            // The frame index used 0 for the oldest samples and increasing indices for newer samples.
            // "Newer" not used now, would have been like mic_data[i][1].
            // But then, since MIC_ARRAY_MAX_FRAME_SIZE_LOG2==0 then there is only one sample per microphone
            //
            for (unsigned i=0; i<NUM_PHYSICAL_MICS; i++) {
                data_mic_samples_delayed [delay_mics_head][i] = latest_frame->data[i][0];
            }

            #if (TEST_SINE_480HZ==0)
                for (unsigned i=0; i<NUM_PHYSICAL_MICS; i++) {
                    // More readable, costs nothing (at least not if -o2)
                    //
                    delay_mics_num_samples_t this_delay =  delay_mics_num_samples[i];
                    unsigned                 this_index = (delay_mics_head - this_delay) % MAX_DELAY_SAMPLE_CNT;
                    //
                    output_to_dac += ((data_mic_samples_delayed [this_index] [i]) >> MIC_ARRAY_NUM_MICS_LOG2);
                }

            #elif (TEST_SINE_480HZ==1) // TODO the accurate value is??
                output_to_dac = sine_table[iof_sine];
                iof_sine = (iof_sine + 1) % SINE_TABLE_SIZE;
            #else
                #error
            #endif

            // Avoids divide/division
            // Avoids dsp_math_divide from lib_dsp, since the xCORE-200 will not divide in one instruction
            //
            output_to_dac =
                    (int) ((((int64_t) output_to_dac * (int64_t) headset_gain)) >> _HEADSET_GAIN_UNITY_BITPOS);

            ch_headset_out <: output_to_dac;
            ch_headset_out <: output_to_dac; // TODO: Why this (it sounds like it's needed)

            delay_mics_head++;
            delay_mics_head %= MAX_DELAY_SAMPLE_CNT;
        }
    }
}
